/**
 * bootprompt.js
 * license: MIT
 * http://github.com/lddubeau/bootprompt
 */
/// <reference types="jquery" />
import "bootstrap";
/** Bootprompt's version */
export declare const VERSION = "6.0.2";
/**
 * The names of the fields that should be defined in a [[Locale]].
 */
export declare type LocaleField = "OK" | "CANCEL" | "CONFIRM";
/** A locale specification. */
export declare type LocaleSpec = Record<LocaleField, string>;
/**
 * Data which we can insert into a DOM document. A string will be interpreted as
 * plain HTML to create DOM nodes.
 */
export declare type DocumentContent = string | Element | DocumentFragment | Text | JQuery;
/**
 * This is the kind of callback all buttons accept.
 *
 * @param this This is set to the jQuery object that models the modal.
 *
 * @param event The jQuery event that triggered the button.
 *
 * @returns ``false`` to keep the modal up. Anything else will hide the modal.
 */
export declare type ButtonCallback = (this: JQuery, event: JQuery.TriggeredEvent) => boolean | void;
/**
 * This interface provides the detailed way to specify a modal button.
 */
export interface Button {
    /**
     * The name of the button shown to the user. The buttons for a modal are
     * specified by a plain JS object with fields whose values conform to the
     * [[Button]] interface. (See [[Buttons]].) If a label is not specified, this
     * fields gets a value from the key with which the [[Button]] object is
     * associated in the map.
     */
    label?: string;
    /** An additional class name to give to the button. */
    className?: string;
    /** A callback to call when the button is clicked. */
    callback?: ButtonCallback;
}
/**
 * A button specification. A button can be specified in one of two ways: by
 * providing a detailed object, or by only providing the button callback and
 * letting Bootprompt use defaults for the rest.
 */
export declare type ButtonSpec = Button | ButtonCallback;
/**
 * We pass an object of this shape to specify all the buttons to be shown in a
 * modal.
 */
export interface Buttons {
    [key: string]: ButtonSpec | undefined;
}
/**
 * A "specialized button" is a button for one of the specialized functions. This
 * interface eliminates the ``callback`` field. The specialized functions
 * overwrite any callback set on these buttons with their own callbacks, so it
 * is not sensible to allow setting ``callback``.
 */
export interface SpecializedButton extends Button {
    callback?: never;
}
/**
 * [[alert]] only takes the ``ok`` button, and so its button specification
 * *must* conform to this interface.
 */
export interface OkButton extends Buttons {
    ok?: SpecializedButton;
}
/**
 * [[confirm]] and [[prompt]] only take the ``confirm`` and ``cancel`` buttons,
 * so their button specification *must* conform to this interface.
 */
export interface ConfirmCancelButtons extends Buttons {
    confirm?: SpecializedButton;
    cancel?: SpecializedButton;
}
/**
 * The set of options which is common to [[alert]], [[confirm]], [[prompt]] and
 * [[dialog]].
 */
export interface CommonOptions {
    /**
     * A message to show in the modal. This is what the modal is asking of the
     * user.
     */
    message?: DocumentContent;
    /**
     * The title of the modal. This appears in the modal header.
     */
    title?: DocumentContent;
    /**
     * Whether to immediately show the modal.
     */
    show?: boolean;
    /**
     * Specifies whether or not to use a backdrop for the modal.
     *
     * - Leaving this option unset is equivalent to setting it to ``"static"``.
     *
     * - ``"static"`` means "display a backdrop that does not affect the modal".
     *
     * - ``true`` means "display a backdrop that closes the modal when clicked on.
     *
     * - ``false`` means don't display a backdrop.
     */
    backdrop?: boolean | "static";
    /**
     * Should the modal have a close button in its header?
     *
     * Default: ``true``.
     */
    closeButton?: boolean;
    /**
     * Should the modal be animated?
     *
     * Default: ``true``.
     */
    animate?: boolean;
    /**
     * An additional class name to add to the modal. This can be a list of space
     * separated names, which will all be added to the name.
     */
    className?: string;
    /**
     * The modal size. Bootstrap provides three sizes: default, large and small.
     *
     * - Not setting this option selects the default size.
     *
     * - ``"large"`` selects the large size.
     *
     * - ``"small"`` selects the small size.
     *
     * **Requires Bootstrap 3.1.0 or newer.**
     */
    size?: "large" | "small";
    /**
     * Which locale to use when generating predefined buttons.
     *
     * Default: ``en``.
     */
    locale?: string;
    /**
     * Whether to swap the button order. The effect varies depending on what
     * function is used:
     *
     * + Since [[alert]] creates a dialog with only one button, it has no visible
     * effect.
     *
     * + For [[confirm]] and [[prompt]] this reverses the order of the
     * cancel/confirm buttons.  The default order is cancel and then confirm. The
     * reverse order is confirm and then cancel.
     *
     * + For [[dialog]] this changes which button is considered "primary". The
     * default is to treat the first button as primary. In reverse, the last
     * button is primary.
     *
     * Default: ``false``.
     */
    swapButtonOrder?: boolean;
    /**
     * Whether to center the modal vertically.
     *
     * **Requires Bootstrap 4.1.0 or newer.**
     *
     * Default: ``false``.
     */
    centerVertical?: boolean;
    /**
     * The container for the modal. The modal HTML markup needs to be put
     * somewhere in the page. This specifies where to put it.
     *
     * Default: ``document.body``.
     */
    container?: string | Element | JQuery;
}
/**
 * This are the options that pertain to the [[dialog]] function.
 */
export interface DialogOptions extends CommonOptions {
    message: DocumentContent;
    /**
     * Specifies what to do if the user hits ``ESC`` on the keyboard.
     *
     * - Leaving this value unset is equivalent to setting it to ``true``.
     *
     * - ``true`` means "dismiss the modal".
     *
     * - ``false`` means "keep the modal displayed".
     *
     * - If a button callback is passed, the callback is called and if it returns
     *   ``false`` the modal remains displayed. Any other value dismisses the
     *   modal.
     */
    onEscape?: boolean | ButtonCallback;
    /**
     * Specifies a callback to call when the user clicks the close button that may
     * be optionally shown in the modal header (when [[closeButton]] is ``true``).
     *
     * If this option is not set, the default behavior is to close the modal.
     *
     * If this option is set, then the modal will remain displayed if the callback
     * returns ``false``. Any other value dismisses the modal.
     */
    onClose?: ButtonCallback;
    /**
     * The list of buttons to show in the dialog.
     */
    buttons?: Buttons;
}
/**
 * Options that are supported by the [[alert]] function.
 */
export interface AlertOptions extends CommonOptions {
    /**
     * A callback for the modal as a whole. This callback is called when the user
     * performs an action that may dismiss the modal.
     *
     * @returns A return value of ``false`` keeps the modal open. Anything else
     * closes it.
     */
    callback?(this: JQuery): boolean | void;
    /** Alerts show only one button. */
    buttons?: OkButton;
    /**
     * Specifies what to do if the user hits ``ESC`` on the keyboard.
     *
     * - Leaving this value unset is equivalent to setting it to ``true``.
     *
     * - ``true`` invoke [[callback]].
     *
     * - ``false`` do not invoke [[callback]].
     *
     * - If a button callback is passed, this callback is called and if it returns
     *   ``false`` [[callback]] is not invoked. Any other value invokes
     *   [[callback]].
     */
    onEscape?: boolean | ButtonCallback;
    /**
     * Specifies a callback to call when the user clicks the close button that may
     * be optionally shown in the modal header (when [[closeButton]] is ``true``).
     *
     * If this option is not set, the default behavior is call [[callback]].
     *
     * If this option is set, this callback is called and if it returns
     * ``false``, [[callback]] is not invoked. Any other value invokes
     * [[callback]].
     *
     * The default value is ``undefined``.
     */
    onClose?: ButtonCallback;
}
/**
 * Options common to the functions that show confirm and cancel buttons
 * ([[confirm]] and [[prompt]]).
 */
export interface ConfirmCancelCommonOptions {
    /** Only the confirm and cancel buttons can be customized. */
    buttons?: ConfirmCancelButtons;
    /**
     * Specifies what to do if the user hits ``ESC`` on the keyboard.
     *
     * - Leaving this value unset is equivalent to setting it to ``true``.
     *
     * - ``true`` invoke [[callback]].
     *
     * - ``false`` do not invoke [[callback]].
     *
     * - If a button callback is passed, this callback is called and if it returns
     *   ``false`` [[callback]] is not invoked. Any other value invokes
     *   [[callback]].
     *
     * If [[callback]] is invoked due to this option, it is invoked with the
     * value ``false``. That is, the user's action is interpreted as a
     * cancellation.
     */
    onEscape?: boolean | ButtonCallback;
    /**
     * Specifies a callback to call when the user clicks the close button that may
     * be optionally shown in the modal header (when [[closeButton]] is ``true``).
     *
     * If this option is not set, the default behavior is call [[callback]].
     *
     * If this option is set, this callback is called and if it returns
     * ``false``, [[callback]] is not invoked. Any other value invokes
     * [[callback]].
     *
     * If [[callback]] is invoked due to this option, it is invoked with the
     * value ``false``. That is, the user's action is interpreted as a
     * cancellation.
     *
     * The default value is ``undefined``.
     */
    onClose?: ButtonCallback;
}
/**
 * The options that are supported by the [[confirm]] function.
 */
export interface ConfirmOptions extends CommonOptions, ConfirmCancelCommonOptions {
    message: DocumentContent;
    /**
     * A callback for the modal as a whole. This callback is called when the user
     * performs an action that may dismiss the modal.
     *
     * @param value ``true`` if the user confirmed, ``false`` otherwise.
     *
     * @returns A return value of ``false`` keeps the modal open. Anything else
     * closes it.
     */
    callback?(this: JQuery, value: boolean): boolean | void;
}
export declare type TextInputType = "text";
export declare type PasswordInputType = "password";
export declare type TextareaInputType = "textarea";
export declare type EmailInputType = "email";
export declare type NumberInputType = "number";
export declare type RangeInputType = "range";
export declare type TimeInputType = "time";
export declare type DateInputType = "date";
export declare type SelectInputType = "select";
export declare type CheckboxInputType = "checkbox";
export declare type RadioInputType = "radio";
/** The input types that present a text field. */
export declare type TextualInputType = TextInputType | PasswordInputType | TextareaInputType | EmailInputType | undefined;
/** The input types that present a numeric field. */
export declare type NumericInputType = NumberInputType | RangeInputType;
/**
 * This type maps an object representing prompt options for a ``select`` input
 * to the value type to be returned by the prompt.
 */
export declare type SelectPromptOptionToValue<T extends {
    inputType: SelectInputType;
    multiple?: boolean;
}> = T["multiple"] extends true ? (string[] | string) : string;
/**
 * This type maps an object representing prompt options to the value type to be
 * returned by the prompt.
 */
export declare type PromptOptionsToValue<T extends {
    inputType?: string;
    multiple?: boolean;
}> = T extends {} ? string : T["inputType"] extends (TextualInputType | NumericInputType | TimeInputType | DateInputType | RadioInputType) ? string : T["inputType"] extends CheckboxInputType ? (string | string[]) : T extends {
    inputType: SelectInputType;
} ? SelectPromptOptionToValue<T> : never;
/**
 * The options that are supported by the [[prompt]] function, irrespective of
 * what ``inputType`` is used.
 */
export interface PromptCommonOptions extends CommonOptions, ConfirmCancelCommonOptions {
    inputType?: string;
    title: DocumentContent;
    /**
     * A callback for the modal as a whole. This callback is called when the user
     * performs an action that may dismiss the modal.
     *
     * @param value ``null`` if the user canceled the modal. Otherwise, it is the
     * value of the ``input`` element.
     *
     * @returns A return value of ``false`` keeps the modal open. Anything else
     * closes it.
     */
    callback?(this: JQuery, value: PromptOptionsToValue<this> | null): boolean | void;
    /**
     * If ``pattern`` is set, the prompt will not close if the input value does
     * not match the pattern. Internally, this option uses the property of the
     * same name on the ``input`` element.
     *
     * It can be used for any input type, but generally only used for ``text``
     * inputs, normally as a fallback for ``email``, ``time``, ``date``,
     * ``number`` or ``range`` inputs where native browser support for those types
     * is lacking.
     *
     * Default: unset, which means that the input is not validated.
     */
    pattern?: string;
}
/**
 * The [[prompt]] options that are supported when the ``inputType`` is one of
 * the textual inputs.
 */
export interface TextPromptOptions extends PromptCommonOptions {
    /**
     * These input types are all textual input types. These input type generate an
     * ``input`` (or ``textarea``) element with the following classes:
     *
     * | ``inputType``  | ``className``                 |
     * | ---------------| ----------------------------- |
     * | ``"text"``     | ``bootprompt-input-text``     |
     * | ``"password"`` | ``bootprompt-input-password`` |
     * | ``"textarea"`` | ``bootprompt-input-textarea`` |
     * | ``"email"``    | ``bootprompt-input-email``    |
     *
     * Default: ``"text"``.
     */
    inputType?: TextualInputType;
    /**
     * An initial value for the input type.
     *
     * Default: ``undefined``.
     */
    value?: string;
    /**
     * Set a maximum length for the input. Users won't be able to type more than
     * the limit established here. Must be a positive numeric value.
     *
     * Default: ``undefined``, which means there's no limit.
     */
    maxlength?: number;
    /**
     * Use this option to provide "helper" text displayed in the text field prior
     * to the user entering anything.
     *
     * There is no limit on the amount of text you may use for your placeholder,
     * but keep in mind that the placeholder disappears when the input either has
     * focus or has a value (depending on the browser). Use the ``message`` option
     * instead of this option to provide help text which you want to always be
     * visible.
     *
     * Default: ``undefined``, which means no placeholder.
     */
    placeholder?: string;
    /**
     * Set this option to ``true`` to require an input value. When ``true``, the
     * prompt will not close if the input value is ``null``, an empty string, or
     * fails the input type's built-in validation constraints.
     *
     * Default: ``undefined``.
     */
    required?: boolean;
}
/**
 * The [[prompt]] options that are supported when the ``inputType`` is one of
 * the numeric inputs.
 */
export interface NumericPromptOptions extends PromptCommonOptions {
    /**
     * These input types are the numeric inputs. These input type generate an
     * ``input`` with the following classes:
     *
     * | ``inputType``  | ``className``              |
     * |----------------|----------------------------|
     * | ``"number"``   | ``bootprompt-input-number``|
     * | ``"range"``    | ``bootprompt-input-range`` |
     *
     * Default: ``"text"``.
     */
    inputType: NumericInputType;
    /**
     * See [[TextPromptOptions.placeholder]].
     */
    placeholder?: string;
    /**
     * An initial value for the input type. Note that even though we are dealing
     * with numeric inputs, the value is a string.
     *
     * **DESIGN NOTE**: If you were to do ``input.value = 1`` directly, the number
     * would automagically be converted to a string by the browser. This automatic
     * conversion of numbers to strings is not formally supported by
     * Bootprompt. The reason is that (again due to the default way the DOM works)
     * when retreiving values from the ``input`` element, the values are always
     * strings. Making the API symmetrical (put a ``number`` in, get a ``number``
     * out) would require more work than reasonable. Note that the DOM
     * specifications do not provide any automagic conversion to help in this
     * endeavor. If you read ``input.value``, it will always be a string quite
     * irrespective of whether you assigned a number to it.
     *
     * Default: ``undefined``, which means no initial value.
     */
    value?: string;
    /**
     * The minimum value allowed by the input.
     *
     * See the MDN article for
     * [``min``](https://developer.mozilla.org/en-US/docs/Web/HTML/Element/input/number#min)
     * for more information.
     *
     * Default: ``undefined``, which means no minimum value.
     */
    min?: string;
    /**
     * The maximum value allowed by the input.
     *
     * See the MDN article for
     * [``max``](https://developer.mozilla.org/en-US/docs/Web/HTML/Element/input/number#max)
     * for more information.
     *
     * Default: ``undefined``, which means no maximum value.
     */
    max?: string;
    /**
     * Can be the string value ``"any"`` (the default), or a positive numeric
     * value.
     *
     * See the
     * [MDN article](https://developer.mozilla.org/en-US/docs/Web/HTML/Element/input/number#step)
     * for more information.
     *
     * Default: ``undefined``, which means ``"any"``.
     */
    step?: string;
    /**
     * See [[TextPromptOptions.required]].
     */
    required?: boolean;
}
/**
 * The [[prompt]] options that are supported when the ``inputType`` is
 * ``"time"``.
 */
export interface TimePromptOptions extends PromptCommonOptions {
    /**
     * An ``inputType`` of value ``"time"`` creates an ``input`` with the class
     * name ``bootprompt-input-time``
     */
    inputType: TimeInputType;
    /**
     * See [[TextPromptOptions.placeholder]].
     */
    placeholder?: string;
    /**
     * An initial value for the input type.
     *
     * A string of the form ``HH:MM:SS``, where ``HH`` can be any zero-padded
     * value between 00 and 23 and ``MM`` and ``SS`` can be any zero-padded number
     * between 00 and 59.
     *
     * Default: ``undefined``, which means no initial value.
     */
    value?: string;
    /**
     * The minimum value allowed by the input. Must be in the same format as
     * [[value]].
     *
     * See the MDN article for
     * [``min``](https://developer.mozilla.org/en-US/docs/Web/HTML/Element/input/time#min)
     * for more information.
     *
     * Default: ``undefined``, which means no minimum value.
     */
    min?: string;
    /**
     * The maximum value allowed by the input. Must be in the same format as
     * [[value]].
     *
     * See the MDN article for
     * [``max``](https://developer.mozilla.org/en-US/docs/Web/HTML/Element/input/time#max)
     * for more information.
     *
     * Default: ``undefined``, which means no maximum value.
     */
    max?: string;
    /**
     * Can be the string value ``"any"`` (the default), or a positive numeric
     * value. For ``"time"`` inputs, this is a value in seconds.
     *
     * See the [MDN
     * article](https://developer.mozilla.org/en-US/docs/Web/HTML/Element/input/time#step)
     * for more information.
     *
     * Default: ``undefined``, which means ``"any"``.
     */
    step?: string;
    /**
     * See [[TextPromptOptions.required]].
     */
    required?: boolean;
}
/**
 * The [[prompt]] options that are supported when the ``inputType`` is
 * ``"date"``.
 */
export interface DatePromptOptions extends PromptCommonOptions {
    /**
     * An ``inputType`` of value ``"date"`` creates an ``input`` with the class
     * name ``bootprompt-input-date``
     */
    inputType: DateInputType;
    /**
     * See [[TextPromptOptions.placeholder]].
     */
    placeholder?: string;
    /**
     * An initial value for the input type.
     *
     * A string of the form ``YYYY-MM-DD``, where ``YYYY`` can be any zero-padded
     * value greater than 0000, ``MM`` can be any zero-padded number between 01
     * and 12, and ``DD`` can be any zero-padded number between 01 and 31.
     *
     * Default: ``undefined``, which means no initial value.
     */
    value?: string;
    /**
     * The minimum value allowed by the input. Must be in the same format as
     * [[value]].
     *
     * See the MDN article for
     * [``min``](https://developer.mozilla.org/en-US/docs/Web/HTML/Element/input/date#min)
     * for more information.
     *
     * Default: ``undefined``, which means no minimum value.
     */
    min?: string;
    /**
     * The maximum value allowed by the input. Must be in the same format as
     * [[value]].
     *
     * See the MDN article for
     * [``max``](https://developer.mozilla.org/en-US/docs/Web/HTML/Element/input/date#max)
     * for more information.
     *
     * Default: ``undefined``, which means no maximum value.
     */
    max?: string;
    /**
     * See [[TextPromptOptions.required]].
     */
    required?: boolean;
}
/**
 * This interface provides options for the ``option`` elements that may be
 * created for ``select``-type inputs.
 */
export interface InputOption {
    /**
     * The text shown to the user for this ``option``.
     */
    text: string;
    /**
     * The value associated with this option.
     */
    value: string;
    /**
     * An optional group to which this option may be assigned. Options having the
     * same ``group`` value are grouped together.
     */
    group?: string;
}
/**
 * The options in common for all ``select``-type inputs.
 */
export interface CommonSelectOptions extends PromptCommonOptions {
    /**
     * An ``inputType`` of value ``"select"`` creates an ``input`` with the class
     * name ``bootprompt-input-select``.
     */
    inputType: SelectInputType;
    /** The list of specifications for the ``option`` elements in this input. */
    inputOptions: InputOption[];
    /**
     * See [[TextPromptOptions.required]].
     */
    required?: boolean;
}
/**
 * The options supported by ``"select"`` inputs that accept multiple values.
 */
export interface MultipleSelectPromptOptions extends CommonSelectOptions {
    /**
     * An initial value for the input type. It is possible to pass an array of
     * values to select multiple initial values.
     *
     * Default: ``undefined``, which means no initial value.
     */
    value?: string[] | string;
    /**
     * This option specifies whether this input accepts multiple values.
     *
     * It must be ``true`` for [[MultipleSelectPromptOptions]].
     */
    multiple: true;
}
/**
 * The options supported by ``"select"`` inputs that accept a single value.
 */
export interface SingleSelectPromptOptions extends CommonSelectOptions {
    /**
     * An initial value for the input type.
     *
     * Default: ``undefined``, which means no initial value.
     */
    value?: string;
    /**
     * This option specifies whether this input accepts multiple values.
     *
     * It must be ``false`` or ``undefined`` for [[SingleSelectPromptOptions]].
     *
     * Default: ``undefined``, which is the same as ``false``.
     */
    multiple?: false;
}
/** All prompt options that create ``select`` inputs. */
export declare type SelectPromptOptions = MultipleSelectPromptOptions | SingleSelectPromptOptions;
/**
 * The options supported by ``"checkbox"`` inputs.
 */
export interface CheckboxPromptOptions extends PromptCommonOptions {
    /**
     * An ``inputType`` of value ``"checkbox"`` creates an ``input`` with the
     * class name ``bootprompt-input-checkbox``.
     */
    inputType: CheckboxInputType;
    /**
     * An initial value for the input type. It is possible to pass an array of
     * values to select multiple initial values.
     *
     * Default: ``undefined``, which means no initial value.
     */
    value?: string | string[];
    /** The list of specifications for the checkboxes to create for this input. */
    inputOptions: InputOption[];
}
/**
 * The options supported by ``"radio"`` inputs.
 */
export interface RadioPromptOptions extends PromptCommonOptions {
    /**
     * An ``inputType`` of value ``"radio"`` creates an ``input`` with the
     * class name ``bootprompt-input-radio``.
     */
    inputType: RadioInputType;
    /**
     * An initial value for the input type.
     *
     * Default: ``undefined``, which means no initial value.
     */
    value?: string;
    /**
     * The list of specifications for the radio buttons to create for this
     * input.
     */
    inputOptions: InputOption[];
}
/**
 * All the options that [[prompt]] accepts.
 */
export declare type PromptOptions = TextPromptOptions | SelectPromptOptions | NumericPromptOptions | TimePromptOptions | DatePromptOptions | CheckboxPromptOptions | RadioPromptOptions;
/**
 * Get all known locales.
 *
 * @returns All known locales.
 */
export declare function locales(): Record<string, LocaleSpec>;
/**
 * Get a single locale.
 *
 * @param name The name of the locale to get.
 *
 * @returns The locale, or ``undefined`` if the locale is unknown.
 */
export declare function locales(name: string): LocaleSpec | undefined;
/**
 * Register a locale.
 *
 * @param name The name of the locale.
 *
 * @param values The locale specification, which determines how to translate
 * each field.
 *
 * @throws {Error} If a field is missing from ``values``.
 */
export declare function addLocale(name: string, values: LocaleSpec): void;
/**
 * Remove a locale. Removing an unknown locale is a no-op.
 *
 * @param name The name of the locale.
 *
 * @throws {Error} If ``name`` is ``"en"``. This locale cannot be removed.
 */
export declare function removeLocale(name: string): void;
/**
 * Set the default locale. Note that this function does not check whether the
 * locale is known.
 */
export declare function setLocale(name: string): void;
/**
 * Set the ``animate`` flag. When the flag is on, the modals produced by this
 * library are animated. When off, they are not animated.
 *
 * **NOTE**: The reason this function exists is to be able to turn off
 * animations during testing. Having the animations turned on can turn simple
 * tests into complicated afairs because it takes a while for a modal to show
 * up or be removed. We do not recommend using this function in production.
 */
export declare function setAnimate(value: boolean): void;
/**
 * Hide all modals created with bootprompt.
 */
export declare function hideAll(): void;
/**
 * This is a general-purpose function allowing to create custom dialogs.
 *
 * @param options The options that govern how the dialog is created.
 *
 * @returns The jQuery object which models the dialog.
 */
export declare function dialog(options: DialogOptions): JQuery;
/**
 * This specialized function provides a dialog similar to the one provided by
 * the DOM ``alert()`` function.
 *
 * **NOTE**: This function is non-blocking, so any code that must happen after
 * the dialog is dismissed should be placed within the callback function for
 * this dialog.
 *
 * @param options The options governing how the dialog is created.
 *
 * @returns A jQuery object that models the dialog.
 */
export declare function alert(options: AlertOptions): JQuery;
/**
 * Specialized function that provides a dialog similar to the one provided by
 * the DOM ``alert()`` function.
 *
 * **NOTE**: This function is non-blocking, so any code that must happen after
 * the dialog is dismissed should be placed within the callback function for
 * this dialog.
 *
 * @param message The message to display.
 *
 * @param callback The callback to call when the dialog has been dismissed.
 *
 * @returns A jQuery object that models the dialog.
 */
export declare function alert(message: string, callback?: AlertOptions["callback"]): JQuery;
/**
 * Specialized function that provides a dialog similar to the one provided by
 * the DOM ``alert()`` function.
 *
 * **NOTE**: This function is non-blocking, so any code that must happen after
 * the dialog is dismissed should await the promise returned by this function.
 *
 * @param messageOrOptions The message to display, or an object specifying the
 * options for the dialog.
 *
 * @returns A promise that resolves once the dialog has been dismissed.
 */
export declare function alert$(messageOrOptions: string | AlertOptions): Promise<void>;
/**
 * This specialized function provides a dialog similar to the one provided by
 * the DOM ``confirm()`` function.
 *
 * **NOTE**: This function is non-blocking, so any code that must happen after
 * the dialog is dismissed should be placed within the callback function for
 * this dialog.
 *
 * @param options The options governing how the dialog is created.
 *
 * @returns A jQuery object that models the dialog.
 */
export declare function confirm(options: ConfirmOptions): JQuery;
/**
 * Specialized function that provides a dialog similar to the one provided by
 * the DOM ``confirm()`` function.
 *
 * **NOTE**: This function is non-blocking, so any code that must happen after
 * the dialog is dismissed should be placed within the callback function for
 * this dialog.
 *
 * @param message The message to display.
 *
 * @param callback The callback to call when the dialog has been dismissed.
 *
 * @returns A jQuery object that models the dialog.
 */
export declare function confirm(message: string, callback: ConfirmOptions["callback"]): JQuery;
/**
 * Specialized function that provides a dialog similar to the one provided by
 * the DOM ``confirm()`` function.
 *
 * **NOTE**: This function is non-blocking, so any code that must happen after
 * the dialog is dismissed should await the promise returned by this function.
 *
 * @param messageOrOptions The message to display, or an object specifying the
 * options for the dialog.
 *
 * @returns A promise that resolves once the dialog has been dismissed.
 */
export declare function confirm$(messageOrOptions: string | ConfirmOptions): Promise<boolean | null>;
/**
 * This specialized function provides a dialog similar to the one provided by
 * the DOM ``prompt()`` function.
 *
 * **NOTE**: This function is non-blocking, so any code that must happen after
 * the dialog is dismissed should be placed within the callback function for
 * this dialog.
 *
 * @param options The options governing how the dialog is created.
 *
 * @returns A jQuery object that models the dialog.
 */
export declare function prompt(options: PromptOptions): JQuery;
/**
 * Specialized function that provides a dialog similar to the one provided by
 * the DOM ``prompt()`` function.
 *
 * **NOTE**: This function is non-blocking, so any code that must happen after
 * the dialog is dismissed should be placed within the callback function for
 * this dialog.
 *
 * @param message The message to display.
 *
 * @param callback The callback to call when the dialog has been dismissed.
 *
 * @returns A jQuery object that models the dialog.
 */
export declare function prompt(message: string, callback: PromptOptions["callback"]): JQuery;
export declare type PromiseValue<T extends PromptOptions> = PromptOptionsToValue<T> | null;
/**
 * Specialized function that provides a dialog similar to the one provided by
 * the DOM ``confirm()`` function.
 *
 * **NOTE**: This function is non-blocking, so any code that must happen after
 * the dialog is dismissed should await the promise returned by this function.
 *
 * @param message The dialog title.
 *
 * @returns A promise that resolves once the dialog has been dismissed.
 */
export declare function prompt$(message: string): Promise<string | null>;
/**
 * Specialized function that provides a dialog similar to the one provided by
 * the DOM ``confirm()`` function.
 *
 * **NOTE**: This function is non-blocking, so any code that must happen after
 * the dialog is dismissed should await the promise returned by this function.
 *
 * @param messageOrOptions An object specifying the options for the dialog.
 *
 * @returns A promise that resolves once the dialog has been dismissed.
 */
export declare function prompt$<T extends PromptOptions>(options: T): Promise<PromiseValue<T>>;
'export as namespace bootprompt;' 
